//#include "Arduino.h"
//#include "SAMDTimerInterrupt.h"
#include "wiring_private.h"
#include "profibus.h"

// uProcessor/Comm. Settings.
#define PORTGROUP_A    0
#define PORTGROUP_B    1
#define RX_PIN              13 //5
#define TX_PIN              8 //6
#define TX_ENABLE_PIN       4 //3
#define SERCOM_RX_PAD       SERCOM_RX_PAD_1
#define UART_TX_PAD         UART_TX_PAD_2

//#define LED_ERROR_ON    PORTB |= _BV (5); // digitalWrite (13, HIGH);
//#define LED_ERROR_OFF   PORTB &= ~_BV (5); // digitalWrite (13, LOW);

#define TX_ENABLE_ON   PORT->Group[PORTGROUP_A].OUTSET.reg = PORT_PA07;  // Enable TX idle state is high
#define TX_ENABLE_OFF  PORT->Group[PORTGROUP_A].OUTCLR.reg = PORT_PA07;  // Enable TX idle state is low

unsigned long samplingtime = 0;
unsigned char uart_buffer[MAX_BUFFER_SIZE];
int16_t  uart_byte_cnt = 0;
int16_t  uart_transmit_cnt = 0;
// Profibus Flags and Variables
int profibus_status;
unsigned char diagnose_status_1;
unsigned char diagnose_status_2;
unsigned char slave_addr;
unsigned char master_addr;
unsigned char Group;
unsigned char Watchdog; //ms
unsigned char new_data;

#if (MAX_OUTPUT_DATA_SIZE > 0)
unsigned char Profibus_out_register[MAX_OUTPUT_DATA_SIZE];       // volatile unsigned char Profibus_out_register[MAX_OUTPUT_DATA_SIZE];
#endif
#if (MAX_INPUT_DATA_SIZE > 0)
unsigned char Profibus_in_register [MAX_INPUT_DATA_SIZE];
#endif
#if (MAX_USER_PARA_SIZE > 0)
unsigned char User_Para[MAX_USER_PARA_SIZE];
#endif
#if (MAX_VENDOR_DATA_SIZE > 0)
unsigned char Vendor_Data[MAX_VENDOR_DATA_SIZE];
#endif
#if (MAX_EXT_DIAG_DATA_SIZE > 0)
unsigned char Diag_Data[MAX_EXT_DIAG_DATA_SIZE];
#endif
unsigned char Input_Data_size;
unsigned char Output_Data_size;
unsigned char User_Para_size;
unsigned char Vendor_Data_size;   // Number of read-in manufacturer-specific bytes

Sercom *sercom = SERCOM1;   // was SERCOM2;
Uart mySerial(&sercom1, RX_PIN, TX_PIN, SERCOM_RX_PAD, UART_TX_PAD);    //Uart mySerial(&sercom0, 8, 10, SERCOM_RX_PAD_1, UART_TX_PAD_2);

///////////////////////////////////////////////////////////////////////////////////////////////////
void setup() 
{
// Initialization
  Serial.begin(BAUD);
  while (!Serial);
   // Pin Mode.
  pinMode(RX_PIN, INPUT);//INPUT_PULLUP);
  pinMode(TX_PIN, OUTPUT); 
  digitalWrite(TX_PIN, LOW);
  pinMode(TX_ENABLE_PIN, OUTPUT);
  // Pin Peripheral.
  pinPeripheral(RX_PIN, PIO_SERCOM);            // PIO_SERCOM    // Assign pin SERCOM functionality
  pinPeripheral(TX_PIN, PIO_SERCOM);            // PIO_SERCOM    // Assign pin SERCOM functionality
//  pinPeripheral(TX_ENABLE_PIN, PIO_SERCOM);   // PIO_SERCOM    // Assign pin SERCOM functionality 

  noInterrupts();           // Disable all interrupts
  init_Profibus();
  setTC3clock();
  interrupts();             // Enable all interrupts
  TX_ENABLE_OFF; // Disable Transmit (Switch to Recieve)
}
///////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////
void loop() 
{
  
}
///////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////
// ============================= FUNCTIONS ============================= //

///////////////////////////////////////////////////////////////////////////////////////////////////
void init_Profibus (void)
{
  unsigned char cnt;
  // Variable init
  profibus_status = PROFIBUS_WAIT_SYN;  // Wait at least Tsyn until allowing RXdata
  diagnose_status_1 = STATION_NOT_READY_;
  diagnose_status_2 = STATUS_2_DEFAULT_; 
  Input_Data_size = 0;
  Output_Data_size = 0;
  User_Para_size = 0;
  Vendor_Data_size = 0;
  Group = 0;

  slave_addr = SLAVE_ADDRESS;  // <<< Temporary address assignment.
  //slave_addr = get_Address();
  // TODO: Read address from EEPROM or switches.
  // Illegal addresses are forced to DEFAULT (126). Set Address can be used to change it.
  if((slave_addr == 0) || (slave_addr > 126))
    slave_addr = DEFAULT_ADD;

  // Clear data
  #if (MAX_OUTPUT_DATA_SIZE > 0)
    memset (Profibus_out_register, 0x00, MAX_OUTPUT_DATA_SIZE);
  #endif
  #if (MAX_INPUT_DATA_SIZE > 0)
    memset (Profibus_in_register, 0x00, MAX_INPUT_DATA_SIZE);
  #endif
  #if (MAX_USER_PARA_SIZE > 0)
    memset (User_Para, 0x00, MAX_USER_PARA_SIZE);  
  #endif
  #if (MAX_VENDOR_DATA_SIZE > 0)
    memset (Vendor_Data, 0x00, MAX_VENDOR_DATA_SIZE);  
  #endif
  #if (MAX_DIAG_DATA_SIZE > 0)
    memset (Diag_Data, 0x00, MAX_DIAG_DATA_SIZE);    
  #endif

  new_data = 0;
  
  PM->APBCMASK.reg |= PM_APBCMASK_SERCOM1;    // Turn on peripheral clock for SERCOM being used.    // TO_DO   Check if necessary.  It is called in the "init"  in wiring.c
  mySerial.begin(BAUD, SERIAL_8E1); 
//  sercom->USART.CTRLB.reg = ~SERCOM_USART_CTRLB_TXEN;
  // Get Synced
//  while (sercom->USART.SYNCBUSY.bit.CTRLB) ;
}
///////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////
void setTC3clock()
{
  GCLK->CLKCTRL.reg = (uint16_t) (GCLK_CLKCTRL_CLKEN | GCLK_CLKCTRL_GEN_GCLK0 | GCLK_CLKCTRL_ID_TCC2_TC3);//GCLK_CLKCTRL_ID( GCM_TCC2_TC3 ));
  while (GCLK->STATUS.bit.SYNCBUSY == 1);              // Wait for synchronization
  TcCount16* TC = (TcCount16*) TC3; // get timer struct
  TC->CTRLA.reg &= ~TC_CTRLA_ENABLE;   // Disable TC
  while (TC->STATUS.bit.SYNCBUSY == 1); // wait for sync
  TC->CTRLA.reg |= TC_CTRLA_MODE_COUNT16;  // Set Timer counter Mode to 16 bits
  while (TC->STATUS.bit.SYNCBUSY == 1); // wait for sync
  TC->CTRLA.reg |= TC_CTRLA_WAVEGEN_NFRQ; // Set TC as normal Normal Frq
  while (TC->STATUS.bit.SYNCBUSY == 1); // wait for sync
  TC->CTRLA.reg |= TC_CTRLA_PRESCALER_DIV1;   // Set perscaler
  while (TC->STATUS.bit.SYNCBUSY == 1); // wait for sync
  
  TC->CC[0].reg = 0xFFF;
  while (TC->STATUS.bit.SYNCBUSY == 1); // wait for sync 
  
  // Interrupts
  TC->INTENSET.reg = 0;              // disable all interrupts
  TC->INTENSET.bit.OVF = 1;          // enable overfollow interrup
  TC->INTENSET.bit.MC0 = 1;          // enable compare match to CC0
  
  // Enable InterruptVector
  NVIC_EnableIRQ(TC3_IRQn);
  // Enable TC
  TC->CTRLA.reg |= TC_CTRLA_ENABLE;
  while (TC->STATUS.bit.SYNCBUSY == 1); // wait for sync
}
///////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////
void resetStartTC3()
{
   TcCount16* TC = (TcCount16*) TC3; // get timer struct
    TC->CTRLBSET.reg |= TC_CTRLBCLR_CMD_RETRIGGER;   // restart
    while (TC->STATUS.bit.SYNCBUSY == 1);
}
///////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////
void stopTC3()
{
  //SerialUSB.println("stp");
  TcCount16* TC = (TcCount16*) TC3; // get timer struct
    TC->CTRLBSET.reg |= TC_CTRLBSET_CMD_STOP;   // Stop counter
    while (TC->STATUS.bit.SYNCBUSY == 1);
}
///////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////
void startTC3()
{
  //SerialUSB.println("sta");
  TcCount16* TC = (TcCount16*) TC3; // get timer struct
    TC->CTRLBSET.reg |= TC_CTRLBSET_CMD_RETRIGGER;   //  Start
    while (TC->STATUS.bit.SYNCBUSY == 1);
}
///////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////
void profibus_RX (void)
{
  unsigned char cnt;
  unsigned char telegramm_type;// identifier SD1,SD2,SD3,SD4
  unsigned char process_data;// Data receive valid format

  // Profibus data types
  unsigned char destination_add;
  unsigned char source_add;
  unsigned char function_code;
  unsigned char FCS_data;   // Frame Check Sequence
  unsigned char PDU_size;   // PDU Size
  unsigned char DSAP_data;  // SAP Destination
  unsigned char SSAP_data;  // SAP Source

  unsigned char DPV1_Read;
  unsigned char Slot_Number;
  unsigned char Index;
  unsigned char Read_req_Lenght;

  //Clear variables
  process_data = false;
  telegramm_type = uart_buffer[0];

  switch (telegramm_type)
  {
        case SD1: // Telegram without data, max. 6 bytes 
            if (uart_byte_cnt != 6) break;
            //uart_buffer[0] = SD1 0x10  // Telegram without data field
            destination_add = uart_buffer[1];
            source_add      = uart_buffer[2];
            function_code   = uart_buffer[3];
            FCS_data        = uart_buffer[4];//(Frame Check Sequence DA + SA + FC) 
            //Check message consistance and end message charter frame ED
            if ((addmatch(destination_add) == false) || (checksum(&uart_buffer[1], 3) != FCS_data) || (uart_buffer[5] != ED)) break;        
            
            process_data = true;
        break;
    
        case SD2: // Telegram with variable length of 4 to 249 bytes and therefore a payload in the range 1 to 246 bytes
                  // Telegram can' t be <4 bytes
            //Serial.println(uart_byte_cnt);          
            if (uart_byte_cnt != uart_buffer[1] + 6) break;
            //uart_buffer[0] = SD2 0x68  // Telegram with variable length
            PDU_size        = uart_buffer[1];
            //uart_buffer[2] = Length (not include in FCS as standard protocol)
            //uart_buffer[3] = Length repeated (not include in FCS as standard protocol)
            destination_add = uart_buffer[4];
            source_add      = uart_buffer[5];
            function_code   = uart_buffer[6];
            FCS_data        = uart_buffer[PDU_size +4];//(Frame Check Sequence DA + SA + FC + PDU) 
            //Check message consistance and end message charter frame ED
            if ((addmatch(destination_add) == false) || (checksum(&uart_buffer[4], PDU_size) != FCS_data) || (uart_buffer[PDU_size +5] != ED)) break;
            //Serial.println("process_data TRUE");
            process_data = true;
        break;
    
        case SD3: // Telegram with 5 bytes data, max. 11 bytes
            if (uart_byte_cnt != 11) break;
            //uart_buffer[0] = SD3 0xA2  // Data telegram fixed
            PDU_size        = 8;              //The PDU has a fixed length of 8 bytes.
            destination_add = uart_buffer[1];
            source_add      = uart_buffer[2];
            function_code   = uart_buffer[3];
            FCS_data        = uart_buffer[9];
            //Check message consistance and end message charter frame ED 
            if ((addmatch(destination_add) == false) || (checksum(&uart_buffer[1], PDU_size) != FCS_data) || (uart_buffer[10] != ED)) break;       
           
            process_data = true;
        break;
    
        case SD4: // Token with 3 Byte Data
            if (uart_byte_cnt != 3) break;
           
            destination_add = uart_buffer[1];
            source_add      = uart_buffer[2];
            //Check only message consistance because no end Charter frame
            if (addmatch(destination_add) == false) break;
        break;
          
  default:         
        process_data = false;
  break;
  }

  // Only evaluate if data is OK
  if (process_data == true)
  {
    master_addr = source_add; // Master address is Source address
    //Service Access Point detected?
    if ((destination_add & 0x80) && (source_add & 0x80))
    {
      DSAP_data = uart_buffer[7];
      SSAP_data = uart_buffer[8];
      // 1) SSAP 62 -> DSAP 60 (Get Diagnostics Request)
      // 2) SSAP 62 -> DSAP 61 (Set Parameters Request)
      // 3) SSAP 62 -> DSAP 62 (Check Config Request)
      // 4) SSAP 62 -> DSAP 60 (Get Diagnostics Request)
      // 5) Data Exchange Request (normal cycle)      
      switch (DSAP_data)
      {
        case DS_REQ: //Actual data request from Master (on line monitoring)
              //(REQUEST_ + FCV_ + SRD_LOW) 5C
              //(REQUEST_ + FCV_ + FCB_ + SRD_LOW) 7C
            
            if (function_code == (REQUEST_ + FCV_ + SRD_LOW) || (REQUEST_ + FCV_ + FCB_ + SRD_LOW))
            {
                  // uart_buffer[7] DSAP_data
                  // uart_buffer[8] SSAP_data
                  // uart_buffer[9] = DPV1_Read;
                  // uart_buffer[10] = Slot_Number;
                  // uart_buffer[11] = Index;
                  // uart_buffer[12] = Read_req_Lenght;
               
                  uart_buffer[12] = 2; // Standard diagnostic 2 bytes
                  uart_buffer[13] = 0x0;
                  uart_buffer[14] = 0x0;
                  // NO STANDARD DIAGNOSTIC INTERRUPT
                  //profibus_send_CMD(SD2, SAP_DISAB, SAP_OFFSET, &uart_buffer[7], 8);   
                  // STANDARD DIAGNOSTIC INTERRUPT
                  profibus_send_CMD(SD2, DATA_LOW, SAP_OFFSET, &uart_buffer[7], 8);
              }         
        break;
        
        case SAP_SET_SLAVE_ADR: // Set Slave Address (SSAP 62 -> DSAP 55)
            // Only possible in the "Wait Parameter" (WPRM) state
            //Telegram to change address has 4 bytes
            //new_addr = uart_buffer[9];
            //IDENT_HIGH_BYTE = uart_buffer[10];
            //IDENT_LOW_BYTE = uart_buffer[11];
            //if (uart_buffer[12] & 0x01) address_Change_lock = TRUE;

            //Note:
            // The PROFIBUS address of a DP slave can be adjusted in two ways:
            //1.Through settings on the device with the help of switches or another control interface.
            //2.By sending a special telegram via the PROFIBUS. Address 126 is reserved as the default address for a DP slave.
            //After the address has been changed, it is necessary to execute a cold-start (power supply off and on again) of the DP slave.
            //uart_buffer[12] bit 0 depend aboud GSD configuration if the device support change address function.
            //GSD PARAMETER -> Set_Slave_Add_supp = 1
 
            // Only accept data for our device
            if ((uart_buffer[10] == IDENT_HIGH_BYTE) && (uart_buffer[11] == IDENT_LOW_BYTE) && (uart_buffer[12] == 0x00))
              {
               //New slave address must store in program memory.(actually this part store new value in local data variable.
               //and the address is not manage to be read from eeprom memory.
               slave_addr = uart_buffer[9];
               //profibus_send_CMD(SC, 0, SAP_OFFSET, &uart_buffer[0], 0);
               profibus_send_CMD(SC, 0, 0, &uart_buffer[0], 0);
               }
        break;

        case SAP_RD_INP: // Master Get Input data (SSAP 62 -> DSAP 56)
          break;


        case SAP_RD_OUTP: // Master Get Output data (SSAP 62 -> DSAP 57)
          break;

        case SAP_GLOBAL_CONTROL: // Global Control Request (SSAP 62 -> DSAP 58)
            // If "Clear Data" high, then PLC CPU on "Stop"
            if (CLEAR_DATA_ == (uart_buffer[9] & CLEAR_DATA_))
            {
              //LED_ERROR_ON;  // Status "PLC not ready"
            }
            else
            {
              //LED_ERROR_OFF; // Status "PLC OK"
            }

            // Calculate Group
            for (cnt = 0;  uart_buffer[10] != 0; cnt++) uart_buffer[10]>>=1;

            // If command is for us
            if (cnt == Group)
            {
              if (UNFREEZE_ == (uart_buffer[9] & UNFREEZE_))
              {
                // Delete FREEZE state
              }
              else if (UNSYNC_ == (uart_buffer[9] & UNSYNC_))
              {
                // Delete SYNC state
              }
              else if (FREEZE_ == (uart_buffer[9] & FREEZE_))
              {
                // Do not read inputs again
              }
              else if (SYNC_ == (uart_buffer[9] & SYNC_))
              {
                // Set outputs only with SYNC command
              }
            }
        break;

        case SAP_GET_CFG: // Get Configuration Request (SSAP 62 -> DSAP 59)

        break;

        case SAP_SLAVE_DIAGNOSIS: // Get Diagnostics Request (SSAP 62 -> DSAP 60)
              // After receiving the diagnosis, the DP slave changes state
              // "Power on Reset" (POR) in the state "Wait Parameter" (WPRM)
              // At the end of initialization ("Data Exchange" state (DXCHG))
              // the master sends a Diagnostics Request a second time to check correct configuration
            //Serial.println("SAP_SLAVE_DIAGNOSIS");
            //REQUEST_ + FCB_ + SRD_HIGH = 6D (first diagnostic request)
            if (function_code == (REQUEST_ + FCB_ + SRD_HIGH))
            {
              // First diagnostic request (call Telegram)

              //uart_buffer[4]  = master_addr;                  // Target Master (with SAP Offset)
              //uart_buffer[5]  = slave_addr + SAP_OFFSET;      // Source Slave (with SAP Offset)
              //uart_buffer[6]  = SLAVE_DATA
              uart_buffer[7]  = SSAP_data;                    // Target SAP Master
              uart_buffer[8]  = DSAP_data;                    // Source SAP Slave
              uart_buffer[9]  = diagnose_status_1;            // Status Byte 1
              uart_buffer[10] = STATUS_2_DEFAULT_ + PRM_REQ_; // Status Byte 2
              uart_buffer[11] = DIAG_SIZE_OK;                 // Status Byte 3
              uart_buffer[12] = MASTER_ADD_DEFAULT;           // Status Byte 4 Master Address
              uart_buffer[13] = IDENT_HIGH_BYTE;              // Ident high
              uart_buffer[14] = IDENT_LOW_BYTE;               // Ident low
              #if (MAX_EXT_DIAG_DATA_SIZE > 0)
              uart_buffer[15] = MAX_EXT_DIAG_DATA_SIZE;           // Device-related diagnosis (number of bytes)
                                
              //for (cnt = 0; cnt < MAX_EXT_DIAG_DATA_SIZE; cnt++)
              //{
              //  uart_buffer[16+cnt] = Diag_Data[cnt];
              //}
              
              memcpy(&uart_buffer[16], &Diag_Data[0], MAX_EXT_DIAG_DATA_SIZE);
     
              #endif
              profibus_send_CMD(SD2, DATA_LOW, SAP_OFFSET, &uart_buffer[7], 8 + MAX_EXT_DIAG_DATA_SIZE);
            }

            //REQUEST_ + FCV_ + SRD_HIGH = 5D (Setting outputs, reading inputs)
            //REQUEST_ + FCV_ + FCB_ + SRD_HIGH = 7D (Setting outputs, reading inputs (toggle)
            else if (function_code == (REQUEST_ + FCV_ + SRD_HIGH) ||
                     function_code == (REQUEST_ + FCV_ + FCB_ + SRD_HIGH))
            {

              // Diagnostic request to check data from Check Config Request
              uart_buffer[7]  = SSAP_data;                    // Target SAP Master
              uart_buffer[8]  = DSAP_data;                    // Source SAP slave
               
              //if (diagnose_status_1 == true)
              //  uart_buffer[9]  = EXT_DIAG_;                  // Status 1
              //else
                  
              uart_buffer[9]  = diagnose_status_1;            // Status 1
              uart_buffer[10] = diagnose_status_2;            // Status 2
              uart_buffer[11] = DIAG_SIZE_OK;                 // Status 3
              uart_buffer[12] = master_addr - SAP_OFFSET;     // Address Master
              uart_buffer[13] = IDENT_HIGH_BYTE;              // Ident high
              uart_buffer[14] = IDENT_LOW_BYTE;               // Ident low
              //uart_buffer[15] = 0x05;     // Device-related diagnostics (number of bytes)
              //uart_buffer[16] = 0x00;     //
              //uart_buffer[17] = 0x20;
              //uart_buffer[18] = 0x20;
              //uart_buffer[19] = 0x00;
              #if (MAX_EXT_DIAG_DATA_SIZE > 0)
              uart_buffer[15] = MAX_EXT_DIAG_DATA_SIZE; // Device-related diagnosis (number of bytes)
                           
              //for (cnt = 0; cnt < MAX_EXT_DIAG_DATA_SIZE; cnt++)
              //{
              //  uart_buffer[16+cnt] = Diag_Data[cnt];
              //}
              
              memcpy(&uart_buffer[16], &Diag_Data[0], MAX_EXT_DIAG_DATA_SIZE);

              profibus_send_CMD(SD2, DATA_LOW, SAP_OFFSET, &uart_buffer[7], 8 + MAX_EXT_DIAG_DATA_SIZE);
              #else
              
              profibus_send_CMD(SD2, DATA_LOW, SAP_OFFSET, &uart_buffer[7], 8);
              #endif
               
            }
        break;

        case SAP_SET_PRM: // Set Parameters Request (SSAP 62 -> DSAP 61)
            // After receiving the parameters, the DP slave changes state "Wait Parameter" (WPRM) in the state "Wait Configuration" (WCFG)
            //uart_buffer[7] SSAP_data
            //uart_buffer[8] DSAP_data
            //uart_buffer[9] Set_Parameter byte 1: Station status
            //uart_buffer[10] Set_Parameter bytes 2 Watchdog factor_1
            //uart_buffer[11] Set_Parameter bytes 3 Watchdog factor_2
            //uart_buffer[12] Set_Parameter byte 4: minimum slave response time (Min. TSDR in tBit)
            //uart_buffer[13] Diagnostic bytes 5 Ident number high byte
            //uart_buffer[14] Diagnostic bytes 6 Ident number low byte
            //uart_buffer[15] Set_Parameter Byte 7: group allocation
            //uart_buffer[16] Set_Parameter Byte 8: DPV1_Status_1 (Note: only if GSD parameter -> DPV1_Slave = 1
            //uart_buffer[17] Set_Parameter Byte 9: DPV1_Status_2
            //uart_buffer[18] Set_Parameter Byte 10: DPV1_Status_3

            // Only accept data for our device
            if ((uart_buffer[13] == IDENT_HIGH_BYTE) && (uart_buffer[14] == IDENT_LOW_BYTE))
            {
              //Set_Parameter byte 1: Station status
              if (ACTIVATE_WATCHDOG_ == (uart_buffer[9] & ACTIVATE_WATCHDOG_))
                 diagnose_status_2 |= WD_ON_;
                //If this bit is set to 0, watchdog monitoring will be disabled. 
              else
               diagnose_status_2 &= ~(WD_ON_); 
              
              if (ACTIVATE_FREEZE_  == (uart_buffer[9] & ACTIVATE_FREEZE_ ))
                //This bit indicates to the DP slave that it is to be operated in Sync-Mode, as soon as the command with the Global_Control function has been transferred.
                //If a DP slave does not support the freeze command, it must set Not_Supported in the diagnostic data. 
                //This request during parameter setting avoids errors when running with payload data. 
                diagnose_status_2 |= FREEZE_MODE_;
              else 
                 diagnose_status_2 &= ~(FREEZE_MODE_); 
                 
              if (ACTIVATE_SYNC_  == (uart_buffer[9] & ACTIVATE_SYNC_ ))
                //This bit indicates to the DP slave that it is to be operated in Freeze-Mode, as soon as the command with the Global_Control function has been transferred.
                //If a DP slave does not support the Sync command, it must set Not_Supported in the diagnostic data. 
                //This request during parameter setting avoids errors when running with payload data. 
                diagnose_status_2 |= SYNC_MODE_;
              else
                diagnose_status_2 &= ~(SYNC_MODE_);
                            
              if (UNLOCK_SLAVE_  == (uart_buffer[9] & UNLOCK_SLAVE_))
              {
                //The DP master sets this bit to 1 when access to a DP slave is to be enabled again for another DP master. This bit has priority over bit 7/ Lock_Req.
              }
              if (LOCK_SLAVE_  == (uart_buffer[9] & LOCK_SLAVE_))
              {
                //The DP master sets this bit to 1 when other masters are to be blocked from accessing a DP slave. 
              }

              //Set_Parameter Byte 8: DPV1_Status_1                
              if (DPV1_MODE_  == (uart_buffer[16] & DPV1_MODE_))
              {
                //The slave should work in DP-V1 mode
              }
              if (FAIL_SAVE_MODE_  == (uart_buffer[16] & FAIL_SAVE_MODE_))
              {
                //The slave should work in Fail_Safe mode
              }
              if (PUBLISHER_MODE_ == (uart_buffer[16] & PUBLISHER_MODE_))
              {
                //The slave should work as Publisher
              }

              if (ACTIVATE_WATCHDOG_ == (uart_buffer[9] & ACTIVATE_WATCHDOG_))
              {
                //note: only if GSD parameter -> WD_Base_1ms_supp = 1
                if (WATCHDOG_TB_1MS == (uart_buffer[16] & WATCHDOG_TB_1MS) && DPV1_MODE_  == (uart_buffer[16] & DPV1_MODE_))
                {
                  //Time base of watchdog is 1 ms (and not 10 ms)
                  //Watchdog = factor_1 * factor_2 * 1 ms
                  Watchdog = uart_buffer[10] * uart_buffer[11];
                }
                else
                {
                  //Watchdog = factor_1 * factor_2 * 10 ms
                  Watchdog = uart_buffer[10] * uart_buffer[11] *10;
                  //Serial.println(Watchdog);
                }  
              }
              else
              {
                //TO CHECK AND COMPLETE
                Watchdog = TIMEOUT_MAX_SDR_TIME / (F_CPU / (1000 * 1000));
              }              

              // User Parameter Size = Length - DA, SA, FC, DSAP, SSAP, 7 Parameter Bytes
              Vendor_Data_size = PDU_size - 12;
              // Read in user parameters
              #if (VENDOR_DATA_SIZE > 0)
              for (cnt = 0; cnt < Vendor_Data_size; cnt++) Vendor_Data[cnt] = uart_buffer[16+cnt];
              #endif
              // Read Group
              for (Group = 0; uart_buffer[15] != 0; Group++) uart_buffer[15]>>=1;
              //profibus_send_CMD(SC, 0, SAP_OFFSET, &uart_buffer[0], 0);
              profibus_send_CMD(SC, 0, 0, &uart_buffer[0], 0);
            }
        break;

        case SAP_CHK_CFG: // Check Config Request (SSAP 62 -> DSAP 62)
        // After receiving the configuration, the DP slave changes state "Wait Configuration" (WCFG) in the "Data Exchange" state (DXCHG)
        // IO configuration:
        // Compact format for max. 16/32 bytes IO , special format for max. 64/132 bytes IO , evaluate several bytes depending on the PDU data size
        // LE / LEr - (DA + SA + FC + DSAP + SSAP) = number of config bytes
        Output_Data_size=0;
        Input_Data_size=0;
            for (cnt = 0; cnt < uart_buffer[1] - 5; cnt++)
            {
              switch (uart_buffer[9+cnt] & CFG_DIRECTION_)
              {
                case CFG_INPUT:
                    Input_Data_size += (uart_buffer[9+cnt] & CFG_BYTE_CNT_) + 1;
                    if (uart_buffer[9+cnt] & CFG_WIDTH_ & CFG_WORD)
                      Input_Data_size += Input_Data_size*2;
                break;

                case CFG_OUTPUT:
                    Output_Data_size += (uart_buffer[9+cnt] & CFG_BYTE_CNT_) + 1;
                    if (uart_buffer[9+cnt] & CFG_WIDTH_ & CFG_WORD)
                      Output_Data_size += Output_Data_size*2;
                break;

                case CFG_INPUT_OUTPUT:
                    Input_Data_size += (uart_buffer[9+cnt] & CFG_BYTE_CNT_) + 1;
                    Output_Data_size += (uart_buffer[9+cnt] & CFG_BYTE_CNT_) + 1;
                    if (uart_buffer[9+cnt] & CFG_WIDTH_ & CFG_WORD)
                    {
                      Input_Data_size += Input_Data_size*2;
                      Output_Data_size += Output_Data_size*2;
                    }
                break;

                case CFG_SPECIAL:
                    // Special format
                    // Manufacturer-specific bytes available?
                    if (uart_buffer[9+cnt] & CFG_SP_VENDOR_CNT_)
                    {
                      // Save the number of manufacturer data
                      Vendor_Data_size = uart_buffer[9+cnt] & CFG_SP_VENDOR_CNT_;
                      // Deduct number of total
                      uart_buffer[1] -= Vendor_Data_size;
                    }

                    // I/O Data
                    switch (uart_buffer[9+cnt] & CFG_SP_DIRECTION_)
                    {
                      case CFG_SP_VOID:
                          // Empty data field
                      break;

                      case CFG_SP_INPUT:
                          Input_Data_size += (uart_buffer[10+cnt] & CFG_SP_BYTE_CNT_) + 1;
                          if (uart_buffer[10+cnt] & CFG_WIDTH_ & CFG_WORD)
                            Input_Data_size += Input_Data_size*2;
                          cnt++;  // We already have this byte
                      break;

                      case CFG_SP_OUTPUT:
                          Output_Data_size += (uart_buffer[10+cnt] & CFG_SP_BYTE_CNT_) + 1;
                          if (uart_buffer[10+cnt] & CFG_WIDTH_ & CFG_WORD)
                            Output_Data_size += Output_Data_size*2;
                          cnt++;  //We already have this byte
                      break;

                      case CFG_SP_INPUT_OPTPUT:
                          // Erst Ausgang...
                          Output_Data_size += (uart_buffer[10+cnt] & CFG_SP_BYTE_CNT_) + 1;
                          if (uart_buffer[10+cnt] & CFG_WIDTH_ & CFG_WORD)
                            Output_Data_size += Output_Data_size*2;
                          // Dann Eingang
                          Input_Data_size += (uart_buffer[11+cnt] & CFG_SP_BYTE_CNT_) + 1;
                          if (uart_buffer[11+cnt] & CFG_WIDTH_ & CFG_WORD)
                            Input_Data_size += Input_Data_size*2;
                          cnt += 2;  // We already have these bytes
                      break;

                    } // Switch End
                    break;

                default:
                    Input_Data_size = 0;
                    Output_Data_size = 0;
                break;
              } // Switch End
            } // For End

            if (Vendor_Data_size != 0)
            {
              // Evaluate
            }
 
            // short acknowledgment
            if ((Input_Data_size >0) && (Input_Data_size <= MAX_INPUT_DATA_SIZE) &&
               (Output_Data_size >0) && (Output_Data_size <= MAX_OUTPUT_DATA_SIZE))
                {
                diagnose_status_1 = STATUS_1_DEFAULT;
                } 
              else
              {
                diagnose_status_1 |= CFG_FAULT_;
              }
                
                //profibus_send_CMD(SC, 0, SAP_OFFSET, &uart_buffer[0], 0);
                profibus_send_CMD(SC, 0, 0, &uart_buffer[0], 0);
        break;

        default:
        // Unknown SAP
        break;
      } //Switch DSAP_data end
    }
    // Destination: Slave address
    else if (destination_add == slave_addr)
    {
      // Status request
      //REQUEST_ + FDL_STATUS 0x49 (status request)
      if (function_code == (REQUEST_ + FDL_STATUS))
      {
        profibus_send_CMD(SD1, FDL_STATUS_OK, 0, &uart_buffer[0], 0);
      }
      // Master sends output data and requests input data(Send and Request Data)
      //REQUEST_ + FCV_ + SRD_HIGH 0x5D (send data request)
      //REQUEST_ + FCV_ + FCB_ + SRD_HIGH 0x7D (Setting outputs, reading inputs (toggle))      
      else if (function_code == (REQUEST_ + FCV_ + SRD_HIGH) ||
               function_code == (REQUEST_ + FCV_ + FCB_ + SRD_HIGH))
      {
        // Read data from master
        #if (MAX_INPUT_DATA_SIZE > 0)
          memcpy(&Profibus_in_register[0], &uart_buffer[7], MAX_INPUT_DATA_SIZE);
          new_data=1;
        #endif

        #if (MAX_OUTPUT_DATA_SIZE > 0)
          memcpy(&uart_buffer[7], &Profibus_out_register[0], MAX_OUTPUT_DATA_SIZE);
        #endif

        #if (MAX_OUTPUT_DATA_SIZE > 0)
        if (diagnose_status_1 != STATUS_1_DEFAULT)
          profibus_send_CMD(SD2, DATA_HIGH, 0, &uart_buffer[7], 0);  // Request a diagnosis
        else
          //profibus_send_CMD(SD2, DATA_LOW, 0, &uart_buffer[7], Input_Data_size);  // send data
          profibus_send_CMD(SD2, DATA_LOW, 0, &uart_buffer[7], Output_Data_size);  // send data
 
       #else if (diagnose_status_1 != STATUS_1_DEFAULT)
          profibus_send_CMD(SD1, DATA_HIGH, 0, &uart_buffer[7], 0);  // Request a diagnosis
        else
          profibus_send_CMD(SC, 0, 0, &uart_buffer[7], 0);          // short acknowledgment
        #endif
      }
    }
  } //Data valid at the end
}
///////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////
void profibus_send_CMD (unsigned char Type,
                        unsigned char function_code,
                        unsigned char sap_offset,
                        unsigned char *pdu,
                        unsigned char length_pdu)
{
  int16_t length_data;

  switch(Type)
  {
        case SD1:
          uart_buffer[0] = SD1;
          uart_buffer[1] = master_addr;
          uart_buffer[2] = slave_addr + sap_offset;
          uart_buffer[3] = function_code;
          uart_buffer[4] = checksum(&uart_buffer[1], 3);
          uart_buffer[5] = ED;
          length_data = 6;
        break;
    
        case SD2:
          uart_buffer[0] = SD2;
          uart_buffer[1] = length_pdu + 3;  // Length of the PDU incl. DA, SA and FC
          uart_buffer[2] = length_pdu + 3;
          uart_buffer[3] = SD2;
          uart_buffer[4] = master_addr;
          uart_buffer[5] = slave_addr + sap_offset;
          uart_buffer[6] = function_code;
          //Data is already filled in before the function is called
          uart_buffer[7+length_pdu] = checksum(&uart_buffer[4], length_pdu + 3);
          uart_buffer[8+length_pdu] = ED;
          length_data = length_pdu + 9;
        break;
    
        case SD3:
          uart_buffer[0] = SD3;
          uart_buffer[1] = master_addr;
          uart_buffer[2] = slave_addr + sap_offset;
          uart_buffer[3] = function_code;
          // Data is already filled in before the function is called
          uart_buffer[9] = checksum(&uart_buffer[4], 8);
          uart_buffer[10] = ED;
          length_data = 11;
        break;
    
        case SD4:
          uart_buffer[0] = SD4;
          uart_buffer[1] = master_addr;
          uart_buffer[2] = slave_addr + sap_offset;
          length_data = 3;
        break;
    
        case SC:
          uart_buffer[0] = SC;
          length_data = 1;
        break;

    default:
      //_NOP();
    break;
  }
  profibus_TX(&uart_buffer[0], length_data);
}
///////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////
void profibus_TX (unsigned char *Data, int16_t Length)
{ 
 //Serial.println("profibus_TX");
  TX_ENABLE_ON;  // Enable Transmit (Switch to Transmit)

//  sercom->USART.CTRLB.reg = SERCOM_USART_CTRLB_TXEN | ~SERCOM_USART_CTRLB_RXEN;
  // Get Synced
//  while (sercom->USART.SYNCBUSY.bit.CTRLB) ;

  profibus_status = PROFIBUS_SEND_DATA;

  uart_byte_cnt = Length;    // Number of bytes to send
  uart_transmit_cnt = 0;     // Payer for sent bytes

//Before was under irq tx
  if (uart_transmit_cnt < uart_byte_cnt)
  {
    for(uart_transmit_cnt = 0; uart_transmit_cnt < uart_byte_cnt; uart_transmit_cnt++)
    {
      Uart_write(Data[uart_transmit_cnt]);
    }
    if (uart_transmit_cnt >= uart_byte_cnt)
    {
      memset (uart_buffer, 0, MAX_BUFFER_SIZE);//clear internal buffer
      //clear uart buffer 
       //while(!sercom->USART.INTFLAG.bit.TXC);
 //     sercom->USART.CTRLB.reg = ~SERCOM_USART_CTRLB_TXEN | SERCOM_USART_CTRLB_RXEN;
      // Get Synced
 //     while (sercom->USART.SYNCBUSY.bit.CTRLB) ;
    }
    resetStartTC3();
  }
}
///////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////
void Uart_write(const unsigned char Data)
{
  //Wait for data register empty and transmission complete
  while(sercom->USART.INTFLAG.bit.DRE == 0 && sercom->USART.INTFLAG.bit.TXC == 0);  // wait for TX data empty
  sercom->USART.DATA.reg = Data;
//  TcCount16* TC = (TcCount16*) TC3; // get timer struct
//  TC->CC[0].reg = TIMEOUT_MAX_TX_TIME;
//  while (TC->STATUS.bit.SYNCBUSY == 1); // wait for sync ; 
}
///////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////
unsigned char checksum(unsigned char *Data, int16_t Length)
{
  unsigned char csum = 0;
  while(Length--)
  {
    csum += Data[Length];
  }
  return csum;
}
///////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////
unsigned char addmatch (unsigned char destination)
{
  if ((destination != slave_addr) &&                // Slave
      (destination != slave_addr + SAP_OFFSET) &&   // SAP Slave
      (destination != BROADCAST_ADD) &&             // Broadcast
      (destination != BROADCAST_ADD + SAP_OFFSET)){ // SAP Broadcast
        return false;
      }
  return true;
}
///////////////////////////////////////////////////////////////////////////////////////////////////

// ============================= INTERRUPTS ============================= //

///////////////////////////////////////////////////////////////////////////////////////////////////
void TC3_Handler()
{  
  TcCount16* TC = (TcCount16*) TC3; // get timer struct
  TC->INTFLAG.bit.OVF = 1;    // writing a one clears the ovf flag
  stopTC3();
  switch (profibus_status)
  {
        case PROFIBUS_WAIT_SYN: // TSYN expired
            //Serial.println("PROFIBUS_WAIT_SYN");
            profibus_status = PROFIBUS_WAIT_DATA;
            TC->CC[0].reg = TIMEOUT_MAX_SDR_TIME;
            while (TC->STATUS.bit.SYNCBUSY == 1); // wait for sync ;
            uart_byte_cnt = 0;
          break;
    
        case PROFIBUS_WAIT_DATA:  // TSDR expired but no data there
          //_NOP();
          break;
    
        case PROFIBUS_GET_DATA:   // TSDR expired and data there
            profibus_status = PROFIBUS_WAIT_SYN;
            // We have already waited TIMEOUT_MAX_RX_TIME of bus idle. So subtract that from Tsyn.
            TC->CC[0].reg = TIMEOUT_MAX_SYN_TIME-TIMEOUT_MAX_RX_TIME;
            while (TC->STATUS.bit.SYNCBUSY == 1); // wait for sync ;
            profibus_RX(); 
          break;
    
        case PROFIBUS_SEND_DATA:  // Transmission timeout expired, go back to reception
            profibus_status = PROFIBUS_WAIT_SYN;
            TC->CC[0].reg = TIMEOUT_MAX_SYN_TIME;
            while (TC->STATUS.bit.SYNCBUSY == 1); // wait for sync ;  
            TX_ENABLE_OFF;  // Disable Transmit (Switch to Recieve)
          break;

    default:
      //_NOP();
    break;
  }
  resetStartTC3();
}
///////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////
//uart receiving RX
void SERCOM1_Handler()
{  
   // RX


  /* Wait for Receive Complete flag */
  while(!sercom->USART.INTFLAG.bit.RXC);                                                                     
  /* Check for errors */
  if (sercom->USART.STATUS.bit.PERR || sercom->USART.STATUS.bit.FERR || sercom->USART.STATUS.bit.BUFOVF)     
  { 
   /* Set the error flag */
   sercom->USART.INTFLAG.bit.ERROR = 1;          
   }
   /* Read data */
   uart_buffer[uart_byte_cnt] = sercom->USART.DATA.reg;                                                  
   // Only read data after Tsyn have expired
   if (profibus_status == PROFIBUS_WAIT_DATA)
   {
    profibus_status = PROFIBUS_GET_DATA;
   }
   // Read data allowed?
   if (profibus_status == PROFIBUS_GET_DATA)
   {
    uart_byte_cnt++;
    // Check for buffer overflow!
    if(uart_byte_cnt >= MAX_BUFFER_SIZE)
       uart_byte_cnt--;
   }
    resetStartTC3();
      





/*     
    mySerial.IrqHandler();
 
    if (mySerial.available())
//    while(sercom->USART.INTFLAG.bit.RXC)
    {
      uart_buffer[uart_byte_cnt] = mySerial.read();
      //uart_buffer[uart_byte_cnt] = sercom->USART.DATA.reg;
      
      //uart_buffer[uart_byte_cnt] = sercom->USART.DATA.reg;
      // Only read data after Tsyn have expired
      if (profibus_status == PROFIBUS_WAIT_DATA)
      {
        profibus_status = PROFIBUS_GET_DATA;
      }
      // Read data allowed?
      if (profibus_status == PROFIBUS_GET_DATA)
      {
        uart_byte_cnt++;
        // Check for buffer overflow!
        if(uart_byte_cnt >= MAX_BUFFER_SIZE)
          uart_byte_cnt--;
      }
        resetStartTC3();
    } 
*/       
}
///////////////////////////////////////////////////////////////////////////////////////////////////
