/*!
 * \file    profibus.h
 * \brief   Ablaufsteuerung Profibus DP-Slave Kommunikation, h-Datei
 * \author  � Joerg S.
 * \date    9.2007 (Erstellung) 9.2009 (Aktueller Stand)
 * \note    Verwendung nur fuer private Zwecke / Only for non-commercial use
 * \note  2010.04.28. Started adaption to AVR. Einar S.
 * \note  2019.08.20. AVR version adapted to Arduino MEGA2560 and Arduino IDE. TadyTheFish
 * \note  2021.09.08 adaption for Atmega328 at 500Mbt/s
 */
 
/*
 * This version works on an Atmega328 and can be compiled with Arduino IDE (i did it with 1.8.13 
 * This slave simply accepts 48 bytes from the Master and returns the 48 bytes to the Master (this is done in void doit() ),
 * and the data is printed on the serial monitor. 
 * 
 * Max Data size = 48 Bytes Input and 48 Bytes Output (96)
 * 
 *  At the moment the Profibus speed is set to 45.45K BAUD if you want to change the speed you need to enable support in the GSD file and uncomment
 *  the define below. You also need to change the DELAY_TBIT to the corresponding Baudrate. 
 *  
 *  If the communication doesn't work try swaping A and B profibus lines. I had this problem with the chinese RS485 boards from Ebay that were labeled wrong
 * 
 * The original version is here: https://www.mikrocontroller.net/topic/106174
 * It was first done on a MSP430F2252 and then ported for MEGA16 (I think) (AVRSoftBus.zip) and then for atXmega32A4U (atXSoftBus.zip)
 * 
 * This is proof of concept, hobby use only!
 * Tested on S7-315 CPU
 * The interface IC for RS485 is MAX485 connected to Serial2. The DE and RE pins of MAX485 are both connected to TX_ENABLE_PIN 2(int0)
 * \note all CPU settings are move in .ino file. Profibus .h contain only profibus declaration not related the type of CPU used.
 */

//#define BAUD 12000000  // DELAY_TBIT = 1 Not working on Nano @16MHz    // 4 at 48Mhz
//#define BAUD 6000000 // DELAY_TBIT = 1 Not working on Nano @16MHz    // 8 at 48Mhz
#define BAUD 3000000  // DELAY_TBIT = 1 Not working on Nano @16MHz    // 16 at 48Mhz
//#define BAUD 1500000  // DELAY_TBIT = 1 Not working on Nano @16MHz    // 32 at 48Mhz
//#define BAUD 500000   // DELAY_TBIT = 4 working on Nano @16MHz        // 96 at 48Mhz
//#define BAUD 187500  // DELAY_TBIT = 11 working on Nano @16MHz        // 256 at 48Mhz
//#define BAUD 93750   // DELAY_TBIT = 21 working on Nano @16MHz        // 512 at 48Mhz
//#define BAUD 45450   // DELAY_TBIT = 44 // Working on Nano @16MHz     // 1056/1057 at 48Mhz
//#define BAUD 31250
//#define BAUD 19200   // DELAY_TBIT = 104 // Working on Nano @16MHz    // 2500 at 48Mhz
//#define BAUD 9600    // DELAY_TBIT = 208 // Working on Nano @16MHz    // 5000 at 48Mhz

#define DELAY_TBIT      16   // This is timer delay for 1 TBIT        // 4966->0 / 4967->95
#define SLAVE_ADDRESS   6   // This is the address of this slave device
//#define TX_ENABLE_PIN   2   // This pin toggles the MAX485 IC from Transmit to Recieve
//#define LED_ERROR_PIN   13  // This is the built in led on pin 13 Arduino Uno - PB5

///////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////
//#define LED_ERROR_ON    PORTB |= _BV (5); // digitalWrite (13, HIGH);
//#define LED_ERROR_OFF   PORTB &= ~_BV (5); // digitalWrite (13, LOW);
//#define LED_ERROR_ON    PORTB |= (1<<5);
//#define LED_ERROR_OFF   PORTB &= ~(1<<5);
//#define LED_ERROR_ON digitalWriteFast(LED_ERROR_PIN,HIGH);
//#define LED_ERROR_OFF digitalWriteFast(LED_ERROR_PIN,LOW);


//#define TX_ENABLE_ON    PORTD |= _BV (2); // digitalWrite (2, HIGH);
//#define TX_ENABLE_OFF   PORTD &= ~_BV (2); // digitalWrite (2, LOW);
//#define TX_ENABLE_ON    PORTD |= (1<<2);//PD2
//#define TX_ENABLE_OFF   PORTD &= ~(1<<2);//PD2
//#define TX_ENABLE_ON    PORTD |= (1<<TX_ENABLE_PIN);
//#define TX_ENABLE_OFF   PORTD &= ~(1<<TX_ENABLE_PIN);
//#define TX_ENABLE_ON  digitalWriteFast(TX_ENABLE_PIN,HIGH);
//#define TX_ENABLE_OFF digitalWriteFast(TX_ENABLE_PIN,LOW);
///////////////////////////////////////////////////////////////////////////////////////////////////
//#define TIMER1_RUN TCCR1B |= (1 << CS11);
//#define TIMER1_STOP TCCR1B &= ~(1 << CS11);
//////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////
// Ident Nummer DP Slave. Arbitrarily chosen. This ID not found in my .GSD library.
///////////////////////////////////////////////////////////////////////////////////////////////////
//#define IDENT_HIGH_BYTE       0xC0
//#define IDENT_LOW_BYTE        0xDE
///////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////
// Addresses
///////////////////////////////////////////////////////////////////////////////////////////////////
#define SAP_OFFSET            128   // Service Access Point Adress Offset
#define BROADCAST_ADD         127
#define DEFAULT_ADD           126   // Delivery address
///////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////
// Command types
///////////////////////////////////////////////////////////////////////////////////////////////////
#define SD1                   0x10  // Telegram without data field
#define SD2                   0x68  // Data telegram variable
#define SD3                   0xA2  // Data telegram fixed
#define SD4                   0xDC  // Token
#define SC                    0xE5  // Short acknowledgment
#define ED                    0x16  // End
///////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////
// Function Codes
///////////////////////////////////////////////////////////////////////////////////////////////////
/* FC Request */
#define REQUEST_              0x40  // SPS: Request Telegramm
#define SRD_HIGH              0x0D  // SPS: Set outputs, read inputs
#define FCV_                  0x10  // SPS: Alternating bit switched on
#define FCB_                  0x20  // SPS: Alternating bit (from frame count)
#define CV_                   0x80  // SPS: Clock Value (Clock synchronization)
#define TE                    0x00  // SPS: Time Event (Clock synchronization)
#define SDA_LOW               0x03  // SPS: Send Data Acknowledged - low priority
#define SDN_LOW               0x04  // SPS: Send Data Not acknowledged - low priority
#define SDA_HIGH              0x05  // SPS: Send Data Acknowledged - high priority
#define SDN_HIGH              0x06  // SPS: Send Data Not acknowledged 
#define MSRD                  0x07  // SPS: Send Request Data with Multicast Reply
#define FDL_STATUS            0x09  // SPS: Request FDL Status
#define SRD_LOW               0x0C  // SPS: Send and Request Data low
#define SRD_HIGH              0x0D  // SPS: Send and Request Data high
#define REQUEST_IDENT         0x0E  // SPS: Request Ident with reply
#define REQUEST_LSAP_STS      0x0F  // SPS: Request LSAP Status with reply 1)

/* FC Response */
#define FDL_STATUS_OK         0x00  // SLA: OK
#define USER_ERR              0x01  // SLA: User error
#define NO_RESORCES           0x02  // SLA: No resources
#define SAP_DISAB             0x03  // SLA: SAP not enabled
#define DATA_LOW              0x08  // SLA: (Data low) Send data inputs
#define NO_RESPONSE           0x09  // SLA: No response data ready
#define DATA_HIGH             0x0A  // SLA: (Data high) Diagnosis pending
#define DATA_LOW_ERR          0x0C  // SLA: Data not received and Data Low
#define DATA_HIGH_ERR         0x0D  // SLA: Data not received and Data High
///////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////
// Service Access Points (DP Slave) MS0
///////////////////////////////////////////////////////////////////////////////////////////////////
#define DS_REQ                51  // Master ask ACT status Diagnostic
#define SAP_Ext_USER_PRM      53  // External parameter, slave responds with SC
#define SAP_SET_SLAVE_ADR     55  // Master sets slave address, slave responds with SC
#define SAP_RD_INP            56  // Master requests input data, slave sends input data
#define SAP_RD_OUTP           57  // Master requests output data, slave sends output data
#define SAP_GLOBAL_CONTROL    58  // Master Control, Slave Does not answer
#define SAP_GET_CFG           59  // Master requests config., Slave sends configuration
#define SAP_SLAVE_DIAGNOSIS   60  // Master requests diagnosis, slave sends diagnosis Daten
#define SAP_SET_PRM           61  // Master sends parameters, slave sends SC
#define SAP_CHK_CFG           62  // Master sends configuration, Slave sends SC
///////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////
// SAP: Global Control (Data Master)
///////////////////////////////////////////////////////////////////////////////////////////////////
#define CLEAR_DATA_           0x02
#define UNFREEZE_             0x04
#define FREEZE_               0x08
#define UNSYNC_               0x10
#define SYNC_                 0x20
///////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////
// SAP: Diagnosis (Answer slave)
// Diagnostics: telegram with min. 6 and max. 244 bytes PDU
///////////////////////////////////////////////////////////////////////////////////////////////////
/* Status Byte 1 */
#define STATUS_1_DEFAULT      0x00
#define STATION_NOT_EXISTENT_ 0x01 //Station does not exist. This bit is always set to zero by a slave. Here a master notices that this slave is not responding. 
#define STATION_NOT_READY_    0x02 //Slave is not ready for data exchange. This bit is set by the DP slave if the DP slave is not yet ready for data exchange. 
#define CFG_FAULT_            0x04 //Fault in the configuration telegram. This bit is set by the DP slave, as soon as the configuration data received most recently from the master does not match that detected by the DP slave.
#define EXT_DIAG_             0x08 //An extended diagnosis follows in the telegram. This bit indicates that further diagnostic blocks follow  starting from byte 7. 
#define NOT_SUPPORTED_        0x10 //Requested function is not supported by slave. This bit is set by a slave, as soon as a function is requested that is not supported by this slave.
#define INV_SLAVE_RESPONSE_   0x20 //Invalid response from slave. This bit is always set to zero by a slave. Here a master notices that a slave gives invalid responses.
#define PRM_FAULT_            0x40 //Fault in parameter telegram. This bit is set by the DP slave if the last parameter telegram was faulty. 
#define MASTER_LOCK           0x80 //Slave has been locked by another master. This bit is always set to zero by a slave. Here a master notices that this slave has been already locked by another master.

/* Status Byte 2 */
#define PRM_REQ_              0x01 //Slave parameters must be reset. If the DP slave sets this bit, its parameters must be reset followed by reconfiguration. The bit remains set until valid parameters have been implemented. 
#define STAT_DIAG_            0x02 //Status diagnostics. If the DP slave sets this bit, the DP master must continue fetching diagnostic data until this bit is deleted again. The DP slave sets this bit when, for example, it is not able to provide any valid user data. 
#define STATUS_2_DEFAULT_     0x04 //Permanently at 1, serves to detect protocol errors.
#define WD_ON_                0x08 //Watchdog on. If this bit is set to 1, watchdog monitoring is enabled.
#define FREEZE_MODE_          0x10 //Freeze command received. This bit is set by the DP slave as soon as this DP slave receives the Freeze command.
#define SYNC_MODE_            0x20 //Sync command received. This bit is set by the DP slave as soon as this DP slave receives the Sync command.
//#define RESERVED_             0x40 //RESERVED
#define DEACTIVATED_          0x80 //Slave is deactivated. This bit is always set to zero by a slave. Here a master notices that this slave has been deactivated and should therefore no longer be controlled cyclically. 

/* Status Byte 3 */
#define DIAG_SIZE_OK          0x00 //Reserved ?????
#define DIAG_SIZE_ERROR       0x80 //If this bit is set, more diagnostic information is available than could be transmitted in one telegram. The master must request further diagnosis. 

/* Status Byte 4 Master Address */
#define MASTER_ADD_DEFAULT    0xFF //Address of master after setting parameters Default is 255 (0xFF)

/* Status Byte 5 Ident Nummer DP Slave */
#define IDENT_HIGH_BYTE       0xC0 //Ident number high byte. Must match GSD congifuration

/* Status Byte 6 Ident Nummer DP Slave */
#define IDENT_LOW_BYTE        0xDE //Ident number low byte. Must match GSD congifuration



/* Extended diagnosis (EXT_DIAG_ = 1) */
#define EXT_DIAG_TYPE_        0xC0  // Bit 6-7 is Diagnostic Type
#define EXT_DIAG_BYTE_CNT_    0x3F  // Bit 0-5 are number of diagnostic bytes
#define EXT_DIAG_DEVICE       0x00  // If bits 7 and 6 = 00, then device-related
#define EXT_DIAG_IDENTIFIER   0x40  // If bits 7 and 6 = 01, then identifier-related
#define EXT_DIAG_CHANNEL      0x80  // If bits 7 and 6 = 10, then channel-related
///////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////
// SAP: Set Parameters Request (Data master)
///////////////////////////////////////////////////////////////////////////////////////////////////
/* Command */
#define ACTIVATE_WATCHDOG_    0x08
#define ACTIVATE_FREEZE_      0x10
#define ACTIVATE_SYNC_        0x20
#define UNLOCK_SLAVE_         0x40  // Slave for other masters released
#define LOCK_SLAVE_           0x80  // Slave locked for other masters

/* DPV1 Status Byte 1 */
#define WATCHDOG_TB_1MS       0x04 //Time base of watchdog is 1 ms (and not 10 ms)
#define PUBLISHER_MODE_       0x20 //The slave should work as Publisher
#define FAIL_SAVE_MODE_       0x40 //The slave should work in Fail_Safe mode
#define DPV1_MODE_            0x80 //The slave should work in DP-V1 mode

/* DPV1 Status Byte 2 */
#define CHECK_CONFIG_MODE_    0x01 //Reduced configuration control
#define UPDATE_ALARM_         0x04 //Switch on update alarm
#define STATUS_ALARM_         0x08 //Switch on status alarm
#define VENDOR_ALARM_         0x10 //Switch on vendor-specific alarm
#define DIAGNOSE_ALARM_       0x20 //Switch on diagnostic alarm
#define PROZESS_ALARM_        0x40 //Switch on process alarm
#define PULL_PLUG_ALARM_      0x80 //Switch on plug alarm (Pull-Plug)

/* DPV1 Status Byte 3 */
#define PARAMETER_BLOCK_      0x08 //Prm_Structure: a parameter block follows
#define ISOCHRON_MODE_ON_     0x10 //Isochronous mode is switched on (Block type = 4)
#define PARAMETER_CMD_ON_     0x80 //Redundancy commands are switched on (Block type = 2)

///////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////
// SAP: Check Config Request (Data Master)
///////////////////////////////////////////////////////////////////////////////////////////////////
#define CFG_DIRECTION_        0x30  // Bit 4-5 is direction. 01 = input, 10 = output, 11 = input / output
#define CFG_INPUT             0x10  // Input
#define CFG_OUTPUT            0x20  // Output
#define CFG_INPUT_OUTPUT      0x30  // Input/Output
#define CFG_SPECIAL           0x00  // Special format if more than 16/32 bytes are to be transferred

#define CFG_KONSISTENZ_       0x80  // Bit 7 is consistency. 0 = byte or word, 1 = over entire module
#define CFG_KONS_BYTE_WORT    0x00  // Byte or word
#define CFG_KONS_MODUL        0x80  // Modul

#define CFG_WIDTH_            0x40  // Bit 6 is IO width. 0 = byte (8bit), 1 = word (16bit)
#define CFG_BYTE              0x00  // Byte
#define CFG_WORD              0x40  // Word

/* Compact format */
#define CFG_BYTE_CNT_         0x0F  // Bit 0-3 are number of bytes or words. 0 = 1 byte, 1 = 2 bytes etc.

/* Special format */
#define CFG_SP_DIRECTION_     0xC0  // Bit 6-7 is direction. 01 = input, 10 = output, 11 = input / output
#define CFG_SP_VOID           0x00  // Empty space
#define CFG_SP_INPUT          0x40  // Input
#define CFG_SP_OUTPUT         0x80  // Output
#define CFG_SP_INPUT_OPTPUT   0xC0  // Input/Output

#define CFG_SP_VENDOR_CNT_    0x0F  // Bits 0-3 are the number of manufacturer-specific bytes. 0 = none

/* Special format / length byte */
#define CFG_SP_BYTE_CNT_      0x3F  // Bit 0-5 are number of bytes or words. 0 = 1 byte, 1 = 2 bytes etc.
///////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////
#define TIMEOUT_MAX_SYN_TIME  33 * DELAY_TBIT // 33 TBit = TSYN
#define TIMEOUT_MAX_RX_TIME   11 * DELAY_TBIT
#define TIMEOUT_MAX_TX_TIME   11 * DELAY_TBIT
#define TIMEOUT_MAX_SDR_TIME  16 * DELAY_TBIT // 15 Tbit = TSDR 187.5    16 Tbit = TSDR 500 & 1.5
///////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////
#define MAX_BUFFER_SIZE        96

#define MAX_INPUT_DATA_SIZE    48    // Number of bytes coming from the master
#define MAX_OUTPUT_DATA_SIZE   48    // Number of bytes going to master
#define MAX_VENDOR_DATA_SIZE    0    // Number of bytes for manufacturer-specific data
#define MAX_EXT_DIAG_DATA_SIZE  0    // Number of bytes for extended diagnostics
//#define MODULE_CNT            1     // Number of modules (input and output modules) for modular station
//#define OUTPUT_MODULE_CNT     1     // Number of output modules
//#define INPUT_MODULE_CNT      1     // Number of input modules
///////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////
// Profibus process control
///////////////////////////////////////////////////////////////////////////////////////////////////
#define PROFIBUS_WAIT_SYN     1
#define PROFIBUS_WAIT_DATA    2
#define PROFIBUS_GET_DATA     3
#define PROFIBUS_SEND_DATA    4
///////////////////////////////////////////////////////////////////////////////////////////////////
